@@ -255,8 +255,9 @@ public:
   ImmutableString() : len(0), base("") {}
   ImmutableString(const char *s, size_t slen)
       : len(slen), base(copystr(s, s + len)) {}
-  ImmutableString(const char *s) : len(strlen(s)), base(copystr(s, s + len)) {}
-  ImmutableString(const std::string &s)
+  explicit ImmutableString(const char *s)
+      : len(strlen(s)), base(copystr(s, s + len)) {}
+  explicit ImmutableString(const std::string &s)
       : len(s.size()), base(copystr(std::begin(s), std::end(s))) {}
   template <typename InputIt>
   ImmutableString(InputIt first, InputIt last)
